package pidgin

  1. Overview
  2. Docs

Module Driver.SexpSource

Dealing with the Sexp format.

Since Csexp shares the same representation as Sexp, the driver supports both.

include SOURCE with type t = Sexp.t
Sourcetype t = Sexp.t

The type that describes the data source.

Sourceval translate_to_pidgin : t -> Repr.t

translate_to_pidgin source must construct a valid Pidgin value (Repr.t) from a source. Error handling is left to the consumer, even though, in an ideal world, we would want to be able to fill any value to Pidgin constructors.

include TARGET with type t := t
Sourceval translate_from_pidgin : Repr.t -> t

translate_from_pidgin repr must construct a valid expression from Repr.t to the target. Error handling is left to the consumer, even though, in an ideal world, we would want to be able to fill any value to the target constructors.